@@ -17,6 +17,7 @@ from .vocab import Vocab
 from .lemmatizer import Lemmatizer
 from .pipeline import DependencyParser, Tensorizer, Tagger, EntityRecognizer
 from .pipeline import SimilarityHook, TextCategorizer, SentenceSegmenter
+from .pipeline import merge_noun_chunks, merge_entities
 from .compat import json_dumps, izip, basestring_
 from .gold import GoldParse
 from .scorer import Scorer
@@ -27,6 +28,7 @@ from .lang.punctuation import TOKENIZER_INFIXES
 from .lang.tokenizer_exceptions import TOKEN_MATCH
 from .lang.tag_map import TAG_MAP
 from .lang.lex_attrs import LEX_ATTRS, is_stop
+from .errors import Errors
 from . import util
 from . import about
 
@@ -105,10 +107,12 @@ class Language(object):
         'similarity': lambda nlp, **cfg: SimilarityHook(nlp.vocab, **cfg),
         'textcat': lambda nlp, **cfg: TextCategorizer(nlp.vocab, **cfg),
         'sbd': lambda nlp, **cfg: SentenceSegmenter(nlp.vocab, **cfg),
-        'sentencizer': lambda nlp, **cfg: SentenceSegmenter(nlp.vocab, **cfg)
+        'sentencizer': lambda nlp, **cfg: SentenceSegmenter(nlp.vocab, **cfg),
+        'merge_noun_chunks': lambda nlp, **cfg: merge_noun_chunks,
+        'merge_entities': lambda nlp, **cfg: merge_entities
     }
 
-    def __init__(self, vocab=True, make_doc=True, meta={}, **kwargs):
+    def __init__(self, vocab=True, make_doc=True, max_length=10**6, meta={}, **kwargs):
         """Initialise a Language object.
 
         vocab (Vocab): A `Vocab` object. If `True`, a vocab is created via
@@ -123,6 +127,15 @@ class Language(object):
             string occurs in both, the component is not loaded.
         meta (dict): Custom meta data for the Language class. Is written to by
             models to add model meta data.
+        max_length (int) :
+            Maximum number of characters in a single text. The current v2 models
+            may run out memory on extremely long texts, due to large internal
+            allocations. You should segment these texts into meaningful units,
+            e.g. paragraphs, subsections etc, before passing them to spaCy.
+            Default maximum length is 1,000,000 characters (1mb). As a rule of
+            thumb, if all pipeline components are enabled, spaCy's default
+            models currently requires roughly 1GB of temporary memory per
+            100,000 characters in one text.
         RETURNS (Language): The newly constructed object.
         """
         self._meta = dict(meta)
@@ -130,12 +143,15 @@ class Language(object):
         if vocab is True:
             factory = self.Defaults.create_vocab
             vocab = factory(self, **meta.get('vocab', {}))
+            if vocab.vectors.name is None:
+                vocab.vectors.name = meta.get('vectors', {}).get('name')
         self.vocab = vocab
         if make_doc is True:
             factory = self.Defaults.create_tokenizer
             make_doc = factory(self, **meta.get('tokenizer', {}))
         self.tokenizer = make_doc
         self.pipeline = []
+        self.max_length = max_length
         self._optimizer = None
 
     @property
@@ -155,7 +171,8 @@ class Language(object):
         self._meta.setdefault('license', '')
         self._meta['vectors'] = {'width': self.vocab.vectors_length,
                                  'vectors': len(self.vocab.vectors),
-                                 'keys': self.vocab.vectors.n_keys}
+                                 'keys': self.vocab.vectors.n_keys,
+                                 'name': self.vocab.vectors.name}
         self._meta['pipeline'] = self.pipe_names
         return self._meta
 
@@ -201,8 +218,7 @@ class Language(object):
         for pipe_name, component in self.pipeline:
             if pipe_name == name:
                 return component
-        msg = "No component '{}' found in pipeline. Available names: {}"
-        raise KeyError(msg.format(name, self.pipe_names))
+        raise KeyError(Errors.E001.format(name=name, opts=self.pipe_names))
 
     def create_pipe(self, name, config=dict()):
         """Create a pipeline component from a factory.
@@ -212,7 +228,7 @@ class Language(object):
         RETURNS (callable): Pipeline component.
         """
         if name not in self.factories:
-            raise KeyError("Can't find factory for '{}'.".format(name))
+            raise KeyError(Errors.E002.format(name=name))
         factory = self.factories[name]
         return factory(self, **config)
 
@@ -237,12 +253,9 @@ class Language(object):
             >>> nlp.add_pipe(component, name='custom_name', last=True)
         """
         if not hasattr(component, '__call__'):
-            msg = ("Not a valid pipeline component. Expected callable, but "
-                   "got {}. ".format(repr(component)))
+            msg = Errors.E003.format(component=repr(component), name=name)
             if isinstance(component, basestring_) and component in self.factories:
-                msg += ("If you meant to add a built-in component, use "
-                        "create_pipe: nlp.add_pipe(nlp.create_pipe('{}'))"
-                        .format(component))
+                msg += Errors.E004.format(component=component)
             raise ValueError(msg)
         if name is None:
             if hasattr(component, 'name'):
@@ -255,11 +268,9 @@ class Language(object):
             else:
                 name = repr(component)
         if name in self.pipe_names:
-            raise ValueError("'{}' already exists in pipeline.".format(name))
+            raise ValueError(Errors.E007.format(name=name, opts=self.pipe_names))
         if sum([bool(before), bool(after), bool(first), bool(last)]) >= 2:
-            msg = ("Invalid constraints. You can only set one of the "
-                   "following: before, after, first, last.")
-            raise ValueError(msg)
+            raise ValueError(Errors.E006)
         pipe = (name, component)
         if last or not any([first, before, after]):
             self.pipeline.append(pipe)
@@ -270,9 +281,8 @@ class Language(object):
         elif after and after in self.pipe_names:
             self.pipeline.insert(self.pipe_names.index(after) + 1, pipe)
         else:
-            msg = "Can't find '{}' in pipeline. Available names: {}"
-            unfound = before or after
-            raise ValueError(msg.format(unfound, self.pipe_names))
+            raise ValueError(Errors.E001.format(name=before or after,
+                                                opts=self.pipe_names))
 
     def has_pipe(self, name):
         """Check if a component name is present in the pipeline. Equivalent to
@@ -290,8 +300,7 @@ class Language(object):
         component (callable): Pipeline component.
         """
         if name not in self.pipe_names:
-            msg = "Can't find '{}' in pipeline. Available names: {}"
-            raise ValueError(msg.format(name, self.pipe_names))
+            raise ValueError(Errors.E001.format(name=name, opts=self.pipe_names))
         self.pipeline[self.pipe_names.index(name)] = (name, component)
 
     def rename_pipe(self, old_name, new_name):
@@ -301,11 +310,9 @@ class Language(object):
         new_name (unicode): New name of the component.
         """
         if old_name not in self.pipe_names:
-            msg = "Can't find '{}' in pipeline. Available names: {}"
-            raise ValueError(msg.format(old_name, self.pipe_names))
+            raise ValueError(Errors.E001.format(name=old_name, opts=self.pipe_names))
         if new_name in self.pipe_names:
-            msg = "'{}' already exists in pipeline. Existing names: {}"
-            raise ValueError(msg.format(new_name, self.pipe_names))
+            raise ValueError(Errors.E007.format(name=new_name, opts=self.pipe_names))
         i = self.pipe_names.index(old_name)
         self.pipeline[i] = (new_name, self.pipeline[i][1])
 
@@ -316,8 +323,7 @@ class Language(object):
         RETURNS (tuple): A `(name, component)` tuple of the removed component.
         """
         if name not in self.pipe_names:
-            msg = "Can't find '{}' in pipeline. Available names: {}"
-            raise ValueError(msg.format(name, self.pipe_names))
+            raise ValueError(Errors.E001.format(name=name, opts=self.pipe_names))
         return self.pipeline.pop(self.pipe_names.index(name))
 
     def __call__(self, text, disable=[]):
@@ -334,11 +340,18 @@ class Language(object):
             >>> tokens[0].text, tokens[0].head.tag_
             ('An', 'NN')
         """
+        if len(text) >= self.max_length:
+            raise ValueError(Errors.E088.format(length=len(text),
+                                                max_length=self.max_length))
         doc = self.make_doc(text)
         for name, proc in self.pipeline:
             if name in disable:
                 continue
+            if not hasattr(proc, '__call__'):
+                raise ValueError(Errors.E003.format(component=type(proc), name=name))
             doc = proc(doc)
+            if doc is None:
+                raise ValueError(Errors.E005.format(name=name))
         return doc
 
     def disable_pipes(self, *names):
@@ -380,8 +393,7 @@ class Language(object):
             >>>            state = nlp.update(docs, golds, sgd=optimizer)
         """
         if len(docs) != len(golds):
-            raise IndexError("Update expects same number of docs and golds "
-                             "Got: %d, %d" % (len(docs), len(golds)))
+            raise IndexError(Errors.E009.format(n_docs=len(docs), n_golds=len(golds)))
         if len(docs) == 0:
             return
         if sgd is None:
@@ -454,6 +466,8 @@ class Language(object):
         else:
             device = None
         link_vectors_to_models(self.vocab)
+        if self.vocab.vectors.data.shape[1]:
+            cfg['pretrained_vectors'] = self.vocab.vectors.name
         if sgd is None:
             sgd = create_default_optimizer(Model.ops)
         self._optimizer = sgd
@@ -461,7 +475,8 @@ class Language(object):
             if hasattr(proc, 'begin_training'):
                 proc.begin_training(get_gold_tuples(),
                                     pipeline=self.pipeline,
-                                    sgd=self._optimizer)
+                                    sgd=self._optimizer,
+                                    **cfg)
         return self._optimizer
 
     def evaluate(self, docs_golds, verbose=False):
@@ -621,9 +636,10 @@ class Language(object):
         """
         path = util.ensure_path(path)
         deserializers = OrderedDict((
-            ('vocab', lambda p: self.vocab.from_disk(p)),
+            ('meta.json', lambda p: self.meta.update(util.read_json(p))),
+            ('vocab', lambda p: (
+                self.vocab.from_disk(p) and _fix_pretrained_vectors_name(self))),
             ('tokenizer', lambda p: self.tokenizer.from_disk(p, vocab=False)),
-            ('meta.json', lambda p: self.meta.update(ujson.load(p.open('r'))))
         ))
         for name, proc in self.pipeline:
             if name in disable:
@@ -666,9 +682,10 @@ class Language(object):
         RETURNS (Language): The `Language` object.
         """
         deserializers = OrderedDict((
-            ('vocab', lambda b: self.vocab.from_bytes(b)),
+            ('meta', lambda b: self.meta.update(ujson.loads(b))),
+            ('vocab', lambda b: (
+                self.vocab.from_bytes(b) and _fix_pretrained_vectors_name(self))),
             ('tokenizer', lambda b: self.tokenizer.from_bytes(b, vocab=False)),
-            ('meta', lambda b: self.meta.update(ujson.loads(b)))
         ))
         for i, (name, proc) in enumerate(self.pipeline):
             if name in disable:
@@ -680,6 +697,27 @@ class Language(object):
         return self
 
 
+def _fix_pretrained_vectors_name(nlp):
+    # TODO: Replace this once we handle vectors consistently as static
+    # data
+    if 'vectors' in nlp.meta and nlp.meta['vectors'].get('name'):
+        nlp.vocab.vectors.name = nlp.meta['vectors']['name']
+    elif not nlp.vocab.vectors.size:
+        nlp.vocab.vectors.name = None
+    elif 'name' in nlp.meta and 'lang' in nlp.meta:
+        vectors_name = '%s_%s.vectors' % (nlp.meta['lang'], nlp.meta['name'])
+        nlp.vocab.vectors.name = vectors_name
+    else:
+        raise ValueError(Errors.E092)
+    if nlp.vocab.vectors.size != 0:
+        link_vectors_to_models(nlp.vocab)
+    for name, proc in nlp.pipeline:
+        if not hasattr(proc, 'cfg'):
+            continue
+        proc.cfg.setdefault('deprecation_fixes', {})
+        proc.cfg['deprecation_fixes']['vectors_name'] = nlp.vocab.vectors.name
+
+
 class DisabledPipes(list):
     """Manager for temporary pipeline disabling."""
     def __init__(self, nlp, *names):
@@ -706,18 +744,11 @@ class DisabledPipes(list):
         if unexpected:
             # Don't change the pipeline if we're raising an error.
             self.nlp.pipeline = current
-            msg = (
-                "Some current components would be lost when restoring "
-                "previous pipeline state. If you added components after "
-                "calling nlp.disable_pipes(), you should remove them "
-                "explicitly with nlp.remove_pipe() before the pipeline is "
-                "restore. Names of the new components: %s"
-            )
-            raise ValueError(msg % unexpected)
+            raise ValueError(Errors.E008.format(names=unexpected))
         self[:] = []
 
 
 def _pipe(func, docs):
     for doc in docs:
-        func(doc)
+        doc = func(doc)
         yield doc
